/****************************************************************
 * backupmailboxes Mark III					*
 *								*
 * $Revision: 2.2 $						*
 * $Date: 2010/02/01 15:49:21 $					*
 *								*
 ****************************************************************/

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <strings.h>
#include <pwd.h>
#include <errno.h>
#include <time.h>

#define DIR_PROC_DELAY 10L

extern int errno;

static char rcsid[] = "$Id: backupinboxes.c,v 2.2 2010/02/01 15:49:21 root Exp $";

static int debug = 0;

#define DEF_CONFIGFILE "/usr/local/admin/etc/inbox-backup.conf"
#define LOCKFILE "/tmp/bkup-inboxes.lck"

#define MAXLINELEN	1024

#define DEF_USERHOME	"/home/users"
#define DEF_BACKUPDIR	"/home/backup-inboxes"
#define DEF_MAXAGE	90*24*60*60 /* 90 days */

struct config {
	char	cfg_userhome[MAXNAMLEN];
	char	cfg_backupdir[MAXNAMLEN];
	time_t	cfg_maxage;
};

int readconfig(struct config *cfg);
int backupfiles(struct config cfg);

int processmaildir(char *maildir, char *backupdir);
char *mungefilename(char *name);
time_t gettimestamp(char *name);

void version(char *prognam);
void usage(char *prognam);

static char configfile[MAXNAMLEN] = DEF_CONFIGFILE;
static time_t start_time;

int main(int argc, char **argv)
{
	struct config cfg;
	int retval = 0, c, fd;
	extern char *optarg;
	extern int optind, optopt;

/*
 * Find the datestamp half an hour ago so that we minimise the number of files we need to look at.
 */
	start_time = time(NULL) - (time_t)(30 * 60);

	while ((c = getopt(argc, argv, "vVd:C:")) != -1)
	{
		switch(c)
		{
			case 'v':
				version(argv[0]);
				break;
			case 'd':
				debug = atoi(optarg);
				break;
			case 'C':
				strncpy(configfile, optarg, MAXNAMLEN - 1);
				configfile[MAXNAMLEN - 1] = NULL;
				break;
			case ':':
				fprintf(stderr, "Option -%c requires an value\n", optopt);
			case 'V':
				version(argv[0]);
			case '?':
				usage(argv[0]);
				retval = -1;
		}
	}

	if (access(LOCKFILE, R_OK) == 0)
	{
		if (debug > 0)
			fprintf(stderr, "Lockfile exists, exiting.\n");
		return -1;
	}

	if ((fd = creat(LOCKFILE, 0600)) != -1)
		close(fd);

	if (retval == 0)
		if ((retval = readconfig(&cfg)) == 0)
			retval = backupfiles(cfg);

	unlink(LOCKFILE);

	return(retval);
}

int readconfig(struct config *cfg)
{
	FILE *fp;
	char buffer[MAXLINELEN + 1], option[10], value[512];
	int i, nval, maxage, retval = 0;

	if ((fp = fopen(configfile, "r")) == NULL)
	{
		strcpy(cfg->cfg_userhome, DEF_USERHOME);
		strcpy(cfg->cfg_backupdir, DEF_BACKUPDIR);
		cfg->cfg_maxage = (time_t) DEF_MAXAGE;
	}
	else
	{
		for (i = 1; fgets(buffer, MAXLINELEN, fp) != NULL ; i++)
		{
			nval = sscanf(buffer, "%9s %511s", option, value);

			if (*option != '#')
			{
				if (nval < 2)
				{
					fprintf(stderr, "Syntax error at line %d in configuration file: only one item on line\n", i);
					retval = -1;
					goto end;
				}

				if (!strcasecmp("userhome", option))
				{
					strcpy(cfg->cfg_userhome, value);
				}
				else if (!strcasecmp("backupdir", option))
				{
					strcpy(cfg->cfg_backupdir, value);
				}
				else if (!strcasecmp("maxage", option))
				{
					if ((maxage = atoi(value)) == 0)
					{
						fprintf(stderr, "Syntax error at line %d in configuration file: non-integer value given for maxage: %s\n", i, value);
						retval = -1;
						goto end;
					}

					cfg->cfg_maxage = maxage * 24 * 60 * 60;
				}
				else
				{
					fprintf(stderr, "Syntax error at line %d in configuration file: unknown keyword: %s\n", i, option);
					retval = -1;
					goto end;
				}
			}
		}
	}
end:
	fclose(fp);

	return retval;
}

int backupfiles(struct config cfg)
{
	int retval = 0;
	char userhome[MAXNAMLEN], backupdir[MAXNAMLEN], maildir[MAXNAMLEN];
	DIR *basedir;
	struct dirent *base_entry;
	struct stat filestatus;
	
	if ((basedir = opendir(cfg.cfg_userhome)) == NULL)
	{
		fprintf(stderr, "backupfiles: Can't open base directory: %s: %s\n", cfg.cfg_userhome, strerror(errno));
		return errno;
	}

	if (debug > 0) fprintf(stderr, "backupfiles: debug: opened directory: %s\n", cfg.cfg_userhome);

	while ((base_entry = readdir(basedir)) != NULL)
	{
		if (debug > 0) fprintf(stderr, "backupfiles: processing: %s\n", base_entry->d_name);

		if (*base_entry->d_name == '.')
			continue;
			
		sprintf(userhome, "%s/%s", cfg.cfg_userhome, base_entry->d_name);

		if (debug > 0) fprintf(stderr, "backupfiles: debug: not a dot file.\n");

		if (stat(userhome, &filestatus) < 0)
		{
			fprintf(stderr, "backupfiles: Can't open get status of user directory: %s: %s\n", userhome, strerror(errno));
			continue;
		}

		if (filestatus.st_uid < 100)
		{
			if (debug > 0)
				fprintf(stderr, "backupfiles: %s: owned by system userid, skipping.\n", userhome);

			continue;
		}

		if (debug > 0) fprintf(stderr, "backupfiles: debug: not owned by system user.\n");

		sprintf(backupdir, "%s/%s", cfg.cfg_backupdir, base_entry->d_name);

		if (stat(backupdir, &filestatus) < 0)
		{
			if (debug > 0) fprintf(stderr, "backupfiles: debug: %s: target directory doesn't exist.\n", backupdir);

			if (!(filestatus.st_mode & S_IFDIR) || (mkdir(backupdir, 0700) < 0))
			{
				fprintf(stderr, "backupfiles: mkdir: %s: %s\n", backupdir, strerror(errno));
				continue;
			}
		}
		
		sprintf(maildir, "%s/Maildir/new", userhome);
		if ((retval = processmaildir(maildir, backupdir)) != 0)

		sprintf(maildir, "%s/Maildir/cur", userhome);
		if ((retval = processmaildir(maildir, backupdir)) != 0)
			fprintf(stderr, "backupfiles: warning: processmaildir returned error code for %s: %s\n", maildir, strerror(retval));
	}
	closedir(basedir);
	
	return 0;
}

int processmaildir(char *maildir, char *backupdir)
{
	int fd;
	char *mungedname, origmsgnam[MAXNAMLEN], bckupmsgnam[MAXNAMLEN];
	DIR *userd;
	time_t filnam_dstamp;
	struct dirent *msg_file;
	struct timespec rqtp = {(time_t)0L, DIR_PROC_DELAY};
	struct timespec rmtp = {(time_t)0L, DIR_PROC_DELAY};

	if ((userd = opendir(maildir)) == NULL)
	{
		fprintf(stderr, "processmaildir: opendir: %s: %s\n", maildir, strerror(errno));
		return errno;
	}

	if (debug > 0) fprintf(stderr, "processmaildir: debug: successfully opened user's maildir: %s\n", maildir);

	while ((msg_file = readdir(userd)) != NULL)
	{
		if (debug > 1) fprintf(stderr, "processmaildir: debug: processing %s.\n", msg_file->d_name);

		if (*msg_file->d_name == '.')
			continue;

		if (debug > 1) fprintf(stderr, "processmaildir: debug: not a dot file.\n");

		mungedname = mungefilename(msg_file->d_name);

		filnam_dstamp = gettimestamp(mungedname);

		if (filnam_dstamp < start_time)
		{
			if (debug > 1) fprintf(stderr, "processmaildir: debug: message %s too old, ignoring.\n", msg_file->d_name);
			continue;
		}

		if (debug > 1) fprintf(stderr, "processmaildir: debug: mundged %s into %s.\n", msg_file->d_name, mungedname);

		sprintf(origmsgnam, "%s/%s", maildir, msg_file->d_name);
		sprintf(bckupmsgnam, "%s/%s", backupdir, mungedname);

		if (access(bckupmsgnam, R_OK) == 0)
		{
			if (debug > 1) fprintf(stderr, "processmaildir: debug: backed up file already exists.\n");

			if ((fd = open(bckupmsgnam, O_RDONLY)) >= 0)
				close(fd);

		}
		else if (link(origmsgnam, bckupmsgnam) < 0)
		{
			if (debug > 1) fprintf(stderr, "processmaildir: debug: no back-up file, linking.\n");

			fprintf(stderr, "processmaildir: linking message: %s: %s\n", msg_file->d_name, strerror(errno));
			continue;
		}
/*		nanosleep(&rqtp, &rmtp); */
	}
	closedir(userd);

	return 0;
}

char *mungefilename(char *orig_name)
{
	static char buffer[MAXNAMLEN];
	char *ptr;

	strcpy(buffer, orig_name);

	if ((ptr = strrchr(buffer, ',')) != NULL)
	{
		*++ptr = 'S';
		*++ptr = NULL;
	}
	else
	{
		fprintf(stderr, "mungefilename: %s: invalid message name.\n", orig_name);
	}

	return buffer;
}

time_t gettimestamp(char *orig_name)
{
	static char buffer[MAXNAMLEN];
	char *ptr;
	time_t value;

	strcpy(buffer, orig_name);

	if ((ptr = strrchr(buffer, '.')) != NULL)
	{
		*ptr = NULL;
		value = (time_t) strtol(buffer, NULL, 10);
	}
	else
	{
		fprintf(stderr, "gettimestamp: %s: invalid message name, doesn't contain a timestamp.\n", orig_name);
	}

	return value;
}

void version(char *prognam)
{
	fprintf(stderr, "%s: %s\n", prognam, rcsid);
}

void usage(char *prognam)
{
	fprintf(stderr, "Usage: %s [-V] [-v] [-C configfile] [-d debuglevel]\n", prognam);
}
